/*	$Id$ */

/*
 * Copyright (c) 1999 LAAS/CNRS                       --  Mon Aug  7 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <portLib.h>

#include <stdlib.h>
#include <h2timeLib.h>

#include "server/stereopixelHeader.h"

#include "pomStruct.h"
#include "libeuler/pomEuler.h"
#include "stereopixelStruct.h"

#include "libstereopixel.h"

#include "cxcore.h"
#include "highgui.h"

/*
 * Globalisms -----------------------------------------------------------
 */

static int maxWidth, maxHeight;		/**< max size of camera images */
static int width, height;		/**< size of camera images */

static char cameraBank[SPIX_PATH_MAX];	/**< current camera bank */
static POSTER_ID cameraPosterId;	/**< current camera poster */
static ViamImageBank *cameraImage;	/**< current poster addr */

static SpixDisparity *disparity;
static Spix3DImage *threeDImage;

static unsigned long imageTicks;
static POM_SENSOR_POS imageTag;	/* for left image */
static unsigned char *source[2];/* buffered sources images */

				/* matrices for reconstruction */
				/* (left camera) */
static double reconstruction[2];
static double rect_to_cam[9];
static double intrinseque[9];
static double intrinseque_rect[9];

static SP_IMAGES_TMP		spixTmpImages[2];	/* for libStereo.a */
static SP_CORRELATION_PARAMS	spixCorrelParams;
static SP_CORRELATION_RESULTS	spixCorrelResults;

DATA_IM3D spixIm3d;

/*
 * Local Functions ------------------------------------------------------
 */

static STATUS	spixLookForImage(const char *bank, int *bilan);
static STATUS	spixCheckPosterSize(int *bilan);

static STATUS	spixUpdateTmpImages(int *bilan);
static STATUS	spixUpdateCorrelResult(int *bilan);
static STATUS	spixUpdateIm3d(int *bilan);

static void	rgbmemcpy(unsigned char *dst, unsigned char *src, size_t n);


/*------------------------------------------------------------------------
 *
 * spixExecTaskInit  --  Codel d'initialisation (SDI/f, ...)
 *
 * Description: Set defaults value
 *
 * Retourne:    OK ou ERROR
 */

STATUS
spixExecTaskInit(int *bilan)
{
   SDI_F->correlParams.win_nb_lig = SPIXDEF_WINLIG;
   SDI_F->correlParams.win_nb_col = SPIXDEF_WINCOL;
   SDI_F->correlParams.disp_min = SPIXDEF_DISPMIN;
   SDI_F->correlParams.disp_max = SPIXDEF_DISPMAX;
   SDI_F->correlParams.correlType = SPIXDEF_CORRELTYPE;
   SDI_F->correlParams.interpolationType = SPIXDEF_INTERPOLATION;
   SDI_F->correlParams.var_th = SPIXDEF_VARTH;
   SDI_F->correlParams.score_th = SPIXDEF_SCORETH;
   SDI_F->correlParams.diff_2_max_th = SPIXDEF_DIFF2MAX;
   SDI_F->correlParams.diff_neighbor_max_th = SPIXDEF_DIFF2NEIGH;
   SDI_F->correlParams.inv_th = SPIXDEF_INVTH;
   spixSetCorrelationParamsExec(&SDI_F->correlParams, bilan);

   SDI_F->blobFilter.enableBlobFilter = SPIXDEF_BLOB;
   SDI_F->blobFilter.sizeMin = SPIXDEF_BLOBSIZEMIN;
   SDI_F->blobFilter.diffMax = SPIXDEF_BLOBDIFFMAX;

   SDI_F->frameCoord = SPIXDEF_FRAME;

   strcpy(SDI_F->saveParams.baseName, SPIXDEF_FRAMEBASENAME);
   SDI_F->saveParams.numDisp = 0;
   SDI_F->saveParams.numDispCV = 0;
   SDI_F->saveParams.num3D = 0;

   spixTmpImages[0].state = SP_IMAGES_TMP_NOT_ALLOCATED;
   spixTmpImages[1].state = SP_IMAGES_TMP_NOT_ALLOCATED;

   spixCorrelResults.state = SP_CORRELATION_RESULTS_NOT_ALLOCATED;

   memset(&spixIm3d, 0, sizeof(DATA_IM3D)); /* sould do what we want */

   cameraBank[0] = '\0';
   width = maxWidth = -1;
   height = maxHeight = -1;
   source[0] = source[1] = NULL;

   return OK;
}


/*------------------------------------------------------------------------
 *
 * spixExecTaskEnd  --  Codel de terminaison
 *
 * Despixription:
 *
 * Retourne:    OK ou ERROR
 */

STATUS
spixExecTaskEnd(void)
{
   int i;

   for(i=0;i<2;i++) {
      if (spixTmpImages[i].state == SP_IMAGES_TMP_ALLOCATED) {
	 free(spixTmpImages[i].intermediaire_moyennes1);
	 free(spixTmpImages[i].intermediaire_moyennes2);
	 free(spixTmpImages[i].im_moyennes);
	 free(spixTmpImages[i].im_sigmas);
      }

      if (source[i] != NULL) free(source[i]);
   }

   if (spixCorrelResults.state == SP_CORRELATION_RESULTS_ALLOCATED)
      sc_clean_correlation_results(&spixCorrelResults);

   empty_data_im3d(&spixIm3d);


   return OK;
}

/*------------------------------------------------------------------------
 * CameraBank
 *
 * Despixription:
 *
 * Bilans:      OK
 */

/* spixCameraBankExec  -  codel EXEC de CameraBank
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
spixCameraBankExec(SpixName *bank, int *bilan)
{
   if (spixLookForImage(bank->id, bilan) != OK) return ETHER;
   if (spixCheckPosterSize(bilan) != OK) return ETHER;

   return ETHER;
}

/*------------------------------------------------------------------------
 * SetCorrelationParams
 *
 * Despixription:
 *
 * Bilans:      OK
 *              S_spixExecTask_PIPO
 */

/* spixSetCorrelationParamsExec  -  codel EXEC de SetCorrelationParams
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
spixSetCorrelationParamsExec(SpixCorrelParams *correlParams, int *bilan)
{
   spixCorrelParams.win_nb_lig = correlParams->win_nb_lig;
   spixCorrelParams.win_nb_col = correlParams->win_nb_col;
   spixCorrelParams.disp_min = correlParams->disp_min;
   spixCorrelParams.disp_max = correlParams->disp_max;
   spixCorrelParams.var_th = correlParams->var_th;
   spixCorrelParams.score_th = correlParams->score_th;
   spixCorrelParams.diff_2_max_th = correlParams->diff_2_max_th;
   spixCorrelParams.diff_neighbor_max_th = correlParams->diff_neighbor_max_th;
   spixCorrelParams.inv_th = correlParams->inv_th;

   switch(correlParams->correlType) {
      case SPIXCT_ZNCC:
	 spixCorrelParams.type_correlation = SP_CORRELATION_ZNCC;
	 break;
      case SPIXCT_ZNCC_SIGMA:
	 spixCorrelParams.type_correlation = SP_CORRELATION_ZNCC_SIGMA;
	 break;
      case SPIXCT_CENSUS:
	 spixCorrelParams.type_correlation = SP_CORRELATION_CENSUS;
	 break;
      case SPIXCT_CENSUS_SIGMA:
	 spixCorrelParams.type_correlation = SP_CORRELATION_CENSUS_SIGMA;
	 break;
   }

   switch(correlParams->interpolationType) {
      case SPIXIT_NO_INTERPOLATION:
	 spixCorrelParams.type_interpolation = SP_INTERPOLATION_DISP_NIL;
	 break;
      case SPIXIT_PARABOLIC_INTERPOLATION:
	 spixCorrelParams.type_interpolation = SP_INTERPOLATION_DISP_PARABOLIQUE_ONLY;
	 break;
      case SPIXIT_BILINEAR_INTERPOLATION:
	 spixCorrelParams.type_interpolation = SP_INTERPOLATION_DISP_BILINERAIRE_ONLY;
	 break;
      case SPIXIT_BOTH_INTERPOLATIONS:
	 spixCorrelParams.type_interpolation = SP_INTERPOLATION_DISP_BOTH;
	 break;
   }

   spixCorrelParams.state = SP_CORRELATION_PARAMS_INITIALIZED;

   if (width > 0 || height > 0)
      if (spixUpdateCorrelResult(bilan) != OK) return ETHER;
   return ETHER;
}

/*------------------------------------------------------------------------
 * Compute
 *
 * Description: Correlation (did you guess it ?)
 */

static int scCorrelStep;

/* spixComputeStart  -  codel START de Spix
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
spixComputeStart(int *bilan)
{
   int i;
   H2TIME date;

   if (posterIoctl(cameraPosterId, FIO_GETDATE, (char *)&date) != OK) {
      *bilan = errnoGet();
      return ETHER;
   }
   imageTicks = date.ntick;

   if (posterTake(cameraPosterId, POSTER_READ) != OK) {
      *bilan = errnoGet();
      return ETHER;
   }

   imageTag = cameraImage->image[0].pos;

   /* copy image pixels -- needs to compute intensity for color images */
   if (cameraImage->image[0].nChannels == 3)
     rgbmemcpy(source[0],
	       &cameraImage->image[0].data[cameraImage->image[0].dataOffset],
	       width*height);
   else
     memcpy(source[0],
	    &cameraImage->image[0].data[cameraImage->image[0].dataOffset],
	    width*height);

   if (cameraImage->image[1].nChannels == 3)
     rgbmemcpy(source[1],
	       &cameraImage->image[1].data[cameraImage->image[1].dataOffset],
	       width*height);
   else
     memcpy(source[1],
	    &cameraImage->image[1].data[cameraImage->image[1].dataOffset],
	    width*height);

   posterGive(cameraPosterId);

   /* reset static structs */
   for(i=0;i<2;i++) {
      spixTmpImages[i].state_intermediaire_moyennes1 = SP_IMAGE_ALLOCATED;
      spixTmpImages[i].state_intermediaire_moyennes2 = SP_IMAGE_ALLOCATED;
   }

   switch (spixCorrelParams.type_correlation) {
      case SPIXCT_ZNCC:
      case SPIXCT_ZNCC_SIGMA:
	 scCorrelStep = 0;
	 break;
      case SPIXCT_CENSUS:
      case SPIXCT_CENSUS_SIGMA:
	 scCorrelStep = 10;
	 break;
   }

   return EXEC;
}

/* spixComputeExec  -  codel EXEC de Spix
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
spixComputeExec(int *bilan)
{
   int ret = 0;
   SP_REPERE_RECONSTRUCTION type_repere=0;
   SP_TYPE_RECONSTRUCTION type_recontruction=0;

   switch (scCorrelStep) {

	/* ZNCC */
      case 0:
	printf ("scCorrelStep = 0\n");
      case 1:
	printf ("scCorrelStep = 1\n");
	 ret = moyennes_sigmas(source[scCorrelStep],
			       &spixTmpImages[scCorrelStep],
			       &spixCorrelParams);
	 break;

      case 2:
	 ret = correlation_zncc(source[0],
				source[1],
				&spixTmpImages[0],
				&spixTmpImages[1],
				&spixCorrelParams, &spixCorrelResults);
	 if (!SDI_F->blobFilter.enableBlobFilter)
	    scCorrelStep++;
	 break;

	 /* CENSUS */
      case 10:
	 ret = correlation_census(source[0],
				  source[1],
				  &spixTmpImages[0].images_sizes,
				  &spixCorrelParams, &spixCorrelResults);
	 if (!SDI_F->blobFilter.enableBlobFilter)
	    scCorrelStep++;
	 break;

	 /* Blob Filter (for both zncc and census) */
      case 3:
      case 11:
	 ret = blob_filter_disparites(&spixCorrelResults,
				      SDI_F->blobFilter.sizeMin,
				      SDI_F->blobFilter.diffMax);
	 break;

	 /* 3D Reconstruction */
      case 4:
      case 12:
	 switch (SDI_F->frameCoord) {
	    case SPIX_RECTIFIED_FRAME:
	       type_repere = SP_REPERE_RECTIFIE;
	       break;
	    case SPIX_LEFT_CAMERA_FRAME:
	       type_repere = SP_REPERE_CAMERA;
	       break;
	 }

	 switch (SDI_F->correlParams.correlType) {
	    case SPIXCT_ZNCC:
	    case SPIXCT_CENSUS:
	       type_recontruction = SP_RECONSTRUCTION_NO_SIGMA;
	       break;

	    case SPIXCT_ZNCC_SIGMA:
	    case SPIXCT_CENSUS_SIGMA:
	       type_recontruction = SP_RECONSTRUCTION_SIGMA;
	       break;
	 }

	 ret = reconstruction_3d(&spixCorrelResults,
				 &spixTmpImages[0].images_sizes,
				 type_repere, type_recontruction,
				 rect_to_cam, reconstruction,
				 intrinseque, intrinseque_rect, &spixIm3d);

	 scCorrelStep = -1;
	 break;

      default:
	 *bilan = S_stereopixel_UNKNOWN_CORREL_STEP;
	 return ETHER;
   }

   scCorrelStep++;

   if (ret != STEREO_OK) {
      *bilan = S_stereopixel_CORRELATION_FAILED;
      return ETHER;
   }

   return (scCorrelStep == 0) ? END : EXEC;
}

/* spixComputeEnd  -  codel END de Spix
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
spixComputeEnd(int *bilan)
{
   DATA_PT3D *pt3d;
   int i,j;

   /* disparity */

   if (posterTake(STEREOPIXEL_DISPARITY_POSTER_ID, POSTER_WRITE) != OK) {
      *bilan = errnoGet();
      return ETHER;
   }

   disparity->imagePos = imageTag;
   disparity->width = width;
   disparity->height = height;
   memcpy(disparity->pixel, spixCorrelResults.im_disparites,
	  width*height*sizeof(float));

   posterGive(STEREOPIXEL_DISPARITY_POSTER_ID);


   /* 3d image */

   if (posterTake(STEREOPIXEL_THREEDIMAGE_POSTER_ID, POSTER_WRITE) != OK) {
      *bilan = errnoGet();
      return ETHER;
   }

   threeDImage->imagePos = imageTag;
   threeDImage->imageTicks = imageTicks;
   threeDImage->width = width;
   threeDImage->height = height;

   for(i=0;i<height;i++) for(j=0;j<width;j++) {
      pt3d = data_im3d_pt(&spixIm3d, i, j);
      if (pt3d->state == GOOD_PT3D) {
	 threeDImage->pixel[i*width+j].good = 1;

	 threeDImage->pixel[i*width+j].x = pt3d->coord_1;
	 threeDImage->pixel[i*width+j].y = pt3d->coord_2;
	 threeDImage->pixel[i*width+j].z = pt3d->coord_3;

	 threeDImage->pixel[i*width+j].sigma = pt3d->covariances_tab[5];
      } else
	 threeDImage->pixel[i*width+j].good = 0;
   }

   posterGive(STEREOPIXEL_THREEDIMAGE_POSTER_ID);
   return ETHER;
}

/* spixComputeInter  -  codel INTER de Spix
   Retourne:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
spixComputeInter(int *bilan)
{
   return ETHER;
}


/* --- SetSaveParams ------------------------------------------------- */

static FILE * spixTestOpen(char *, char *, int, int *);

/* spixSetSaveParamsExec  -  codel EXEC de SetSaveParams
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
spixSetSaveParamsExec(SpixSaveParams *saveParams, int *bilan)
{
   FILE *f;

   f = spixTestOpen(saveParams->baseName, "disparity", saveParams->numDisp,
		    bilan);
   if (f) fclose(f);

   f = spixTestOpen(saveParams->baseName, "im3d", saveParams->num3D,
		    bilan);
   if (f) fclose(f);

   return ETHER;
}

static FILE *
spixTestOpen(char *base, char *file, int num, int *bilan)
{
   FILE *f;

   sprintf(SDI_F->fileName, "%s/%s.%04d", base, file, num);

   f = fopen(SDI_F->fileName, "w");
   if (f == NULL) {
      fprintf(stderr,
	      "spixSaveParams: cannot open %s\n", SDI_F->fileName);
      *bilan = S_stereopixel_CANNOT_OPEN_FILE;
      return NULL;
   }

   return f;
}


/*------------------------------------------------------------------------
 * SaveDisparity
 *
 * Description:
 */

/* spixSaveDisparityExec  -  codel EXEC de SaveDisparity
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
spixSaveDisparityExec(int *bilan)
{
  FILE *f;
  int size, v, i;
  float min, *p;

  f = spixTestOpen(SDI_F->saveParams.baseName, "disparity",
		   SDI_F->saveParams.numDisp, bilan);
  if (*bilan != OK) return ETHER;

  if (posterTake(STEREOPIXEL_DISPARITY_POSTER_ID, POSTER_READ) != OK) {
    *bilan = errnoGet();
    return ETHER;
  }

  /* get min (disparity is negative) */
  size = disparity->width * disparity->height;
  min = *disparity->pixel;
  for(p = disparity->pixel, i = 0; i < size ; p++, i++)
    if (min > *p) min = *p;

  /* make a grey-level ppm */
  /* THOMAS: invert width and height so that images are viewable (pb with old calife float images) */
  fprintf(f, "P5 %u %u %u\n", disparity->height, disparity->width, (1<<16)-1);
  for(p = disparity->pixel, i = 0; i < size ; p++, i++) {
    v = (*p) / min * ((1<<16)-1);
    fputc(v >> 8, f);
    fputc(v & 0xff, f);
  }

  fclose(f);

  sprintf(SDI_F->fileName, "%s/%s.pos.%04d",
	  SDI_F->saveParams.baseName, "disparity",
	  SDI_F->saveParams.numDisp);
  pomWriteSensorPos(SDI_F->fileName, &disparity->imagePos);

  posterGive(STEREOPIXEL_DISPARITY_POSTER_ID);
  SDI_F->saveParams.numDisp++;
  return ETHER;
}

/*------------------------------------------------------------------------
 * CVDumpDisparity
 *
 * Description:
 */

/* spixSaveDisparityExec  -  codel EXEC de SaveDisparity
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
spixCVDumpDisparityExec(SpixName *cvDisparityFile, int *bilan)
{
  int x, y;
  float *p;
  CvSize cvImageSize;
  IplImage* cvDisparityImage;

  /* create the openCV image */
  cvImageSize = cvSize(disparity->width, disparity->height);
  cvDisparityImage = cvCreateImage(cvImageSize, IPL_DEPTH_32F , 1);

  if (posterTake(STEREOPIXEL_DISPARITY_POSTER_ID, POSTER_READ) != OK) {
    *bilan = errnoGet();
    return ETHER;
  }

  /* fill the openCV image */
  p = disparity->pixel;

  for(x = 0; x < cvImageSize.width ; ++x) {
    for(y = 0; y < cvImageSize.height ; ++y) {
      ((float*)(cvDisparityImage->imageData + cvDisparityImage->widthStep*y))[x] = *p;
      ++p;
    }
  }

  sprintf(SDI_F->fileName, "%s/%s.%04d.xml",
	  SDI_F->saveParams.baseName, cvDisparityFile->id, SDI_F->saveParams.numDispCV);
  printf ("File saved in : %s\n", SDI_F->fileName);
  cvSave(SDI_F->fileName, cvDisparityImage, NULL, NULL, cvAttrList(0,0));
  printf("Disparity image saved in %s\n",SDI_F->fileName);

  IplImage* img = cvCreateImage(cvImageSize, IPL_DEPTH_8U, 1);
  double dispMin, dispMax;
  cvMinMaxLoc(cvDisparityImage, &dispMin, &dispMax, NULL, NULL, NULL);
  cvConvertScale(cvDisparityImage, img, 255.0/dispMin, 0);
  sprintf(SDI_F->fileName, "%s/%s.%04d.tiff",
	  SDI_F->saveParams.baseName, cvDisparityFile->id, SDI_F->saveParams.numDispCV);
  cvSaveImage(SDI_F->fileName, img);
  printf("Disparity image saved in %s\n",SDI_F->fileName);

  cvReleaseImage(&img);

  cvRelease(&cvDisparityImage);

  sprintf(SDI_F->fileName, "%s/%s.pos.%04d",
	  SDI_F->saveParams.baseName, cvDisparityFile->id, SDI_F->saveParams.numDispCV);
  pomWriteSensorPos(SDI_F->fileName, &disparity->imagePos);

  posterGive(STEREOPIXEL_DISPARITY_POSTER_ID);
  SDI_F->saveParams.numDispCV++; /* Increment the file number */
  return ETHER;
}

/*------------------------------------------------------------------------
 * Save3DImage
 *
 * Description:
 */

/* spixSave3DImageExec  -  codel EXEC de Save3DImage
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
spixSave3DImageExec(int *bilan)
{
   spixTestOpen(SDI_F->saveParams.baseName, "im3d",
		   SDI_F->saveParams.num3D, bilan);
   if (*bilan != OK) return ETHER;

   strcat(SDI_F->fileName, ".gz");
   if (write_data_im3d_ascii(&spixIm3d, SDI_F->fileName) != ERR_IM3D_OK) {
      *bilan = S_stereopixel_CANNOT_OPEN_FILE;
      return ETHER;
   }

   if (posterTake(STEREOPIXEL_THREEDIMAGE_POSTER_ID, POSTER_READ) != OK) {
      *bilan = errnoGet();
      return ETHER;
   }

   sprintf(SDI_F->fileName, "%s/%s.pos.%04d",
	   SDI_F->saveParams.baseName, "im3d", SDI_F->saveParams.num3D);
   pomWriteSensorPos(SDI_F->fileName, &threeDImage->imagePos);

   posterGive(STEREOPIXEL_THREEDIMAGE_POSTER_ID);

   SDI_F->saveParams.num3D++;
   return ETHER;
}

/*----------------------------------------------------------------------
 *
 * spixDoNotCreatePoster
 */

STATUS
spixDoNotCreatePoster(char *name, int size, POSTER_ID *pPosterId)
{
   return OK;
}


/*
 * ----------------------------------------------------------------------
 * Local functions
 */

static STATUS
spixLookForImage(const char *bank, int *bilan)
{
   if (!strcmp(cameraBank,bank)) return OK;

   sprintf(cameraBank, "viam%s", bank);

   if (posterFind(cameraBank, &cameraPosterId) != OK) {
      fprintf(stderr, "spix: cannot find %s\n", SDI_F->fileName);
      *bilan = S_stereopixel_NO_CAMERA_POSTER;
      return ERROR;
   }

   if (posterTake(cameraPosterId, POSTER_READ) != OK) {
      *bilan = errnoGet();
      return ERROR;
   }

   cameraImage = posterAddr(cameraPosterId);
   if (cameraImage == NULL) {
      *bilan = S_stereopixel_NO_CAMERA_POSTER;
      return ERROR;
   }

   if (cameraImage->nImages != 2) {
      posterGive(cameraPosterId);
      *bilan = S_stereopixel_BANK_IS_NOT_STEREO;
      return ERROR;
   }

   /* Gets image size ------------------------------------------------- */

   width = cameraImage->image[0].width;
   height = cameraImage->image[0].height;

   /* Matrices -------------------------------------------------------- */

   /* XXX will cleanup this soon */
   reconstruction[0] = cameraImage->calibration.pbaseline;
   reconstruction[1] = cameraImage->calibration.baseline;

   memcpy(rect_to_cam,
	  cameraImage->image[0].calibration.rotation,
	  sizeof(rect_to_cam));
   memcpy(intrinseque,
	  cameraImage->image[0].calibration.intrinsic,
	  sizeof(intrinseque));
   memcpy(intrinseque_rect,
	  cameraImage->image[0].calibration.intrirect,
	  sizeof(intrinseque_rect));

   posterGive(cameraPosterId);

   fprintf(stderr, "spix: using %dx%d images\n", width, height);
   return OK;
}


/* --- spixCheckPosterSize ------------------------------------------- */

static STATUS
spixCheckPosterSize(int *bilan)
{
   int i;
   size_t size;

   if (maxWidth >= width && maxHeight >= height) return OK;

   /* create disparity */

   if (STEREOPIXEL_DISPARITY_POSTER_ID != NULL) {
      posterDelete(STEREOPIXEL_DISPARITY_POSTER_ID);
      fprintf(stderr, "spix: destroyed %s poster\n",
	      STEREOPIXEL_DISPARITY_POSTER_NAME);
   }

   size = (char *)&disparity->pixel[width*height] - (char *)disparity;
   if (posterCreate(STEREOPIXEL_DISPARITY_POSTER_NAME,
		    size, &STEREOPIXEL_DISPARITY_POSTER_ID) != OK) {
      fprintf(stderr, "Could not create %s poster\n",
	      STEREOPIXEL_DISPARITY_POSTER_NAME);
      *bilan = S_stereopixel_CANNOT_CREATE_POSTER;
      return ERROR;
   } else {
      fprintf(stderr, "spix: created %s poster\n",
	      STEREOPIXEL_DISPARITY_POSTER_NAME);
   }

   disparity = posterAddr(STEREOPIXEL_DISPARITY_POSTER_ID);

   /* initialize sizes */
   if (posterTake(STEREOPIXEL_DISPARITY_POSTER_ID, POSTER_WRITE) != OK) {
      *bilan = errnoGet();
      return ETHER;
   }

   disparity->width = width;
   disparity->height = height;
   posterGive(STEREOPIXEL_DISPARITY_POSTER_ID);

   /* create 3D Image */

   if (STEREOPIXEL_THREEDIMAGE_POSTER_ID != NULL) {
      posterDelete(STEREOPIXEL_THREEDIMAGE_POSTER_ID);
      fprintf(stderr, "spix: destroyed %s poster\n",
	      STEREOPIXEL_DISPARITY_POSTER_NAME);
   }

   size = (char *)&threeDImage->pixel[width*height] - (char *)threeDImage;
   if (posterCreate(STEREOPIXEL_3DIMAGE_POSTER_NAME,
		    size, &STEREOPIXEL_THREEDIMAGE_POSTER_ID) != OK) {
      fprintf(stderr, "Could not create %s poster\n",
	      STEREOPIXEL_3DIMAGE_POSTER_NAME);
      *bilan = S_stereopixel_CANNOT_CREATE_POSTER;
      return ERROR;
   } else {
      fprintf(stderr, "spix: created %s poster\n",
	      STEREOPIXEL_3DIMAGE_POSTER_NAME);
   }

   threeDImage = posterAddr(STEREOPIXEL_THREEDIMAGE_POSTER_ID);

   /* initialize size */
   if (posterTake(STEREOPIXEL_THREEDIMAGE_POSTER_ID, POSTER_WRITE) != OK) {
      *bilan = errnoGet();
      return ETHER;
   }

   threeDImage->width = width;
   threeDImage->height = height;

   posterGive(STEREOPIXEL_THREEDIMAGE_POSTER_ID);

   /* initialization of static structs */
   if (spixUpdateTmpImages(bilan) != OK ||
       spixUpdateCorrelResult(bilan) != OK ||
       spixUpdateIm3d(bilan) != OK)
      return ERROR;

   /* allocation of internal images */
   for(i=0;i<2;i++) {
      if (source[i] != NULL) free(source[i]);

      source[i] = malloc(width*height);
      if (source[i] == NULL) {
	 *bilan = S_stereopixel_NOT_ENOUGH_MEMORY;
	 return ERROR;
      }
   }

   maxWidth = width;
   maxHeight = height;
   return OK;
}

static STATUS
spixUpdateTmpImages(int *bilan)
{
   int i;

   for(i=0;i<2;i++) {

      if (spixTmpImages[i].state == SP_IMAGES_TMP_ALLOCATED) {
	 free(spixTmpImages[i].intermediaire_moyennes1);
	 free(spixTmpImages[i].intermediaire_moyennes2);
	 free(spixTmpImages[i].im_moyennes);
	 free(spixTmpImages[i].im_sigmas);
      }

      /*
       * Le  images_size , un poil truand et/ou pipot, mais bon...
       */

      /* En gros, tout est faux, mais a fait ce qu'on veut */
      spixTmpImages[i].images_sizes.previousReduction = 1;
      spixTmpImages[i].images_sizes.i_min = 0;
      spixTmpImages[i].images_sizes.i_max = height;
      spixTmpImages[i].images_sizes.j_min = 0;
      spixTmpImages[i].images_sizes.j_max = width;

      spixTmpImages[i].images_sizes.taux_reduction_i = 1;
      spixTmpImages[i].images_sizes.taux_reduction_j = 1;

      spixTmpImages[i].images_sizes.nb_lig_final = height;
      spixTmpImages[i].images_sizes.nb_col_final = width;

      /*
       * les 2 images intermdiaires pour le cacul des moyennes et sigmas
       */

      spixTmpImages[i].intermediaire_moyennes1 =
	malloc(width*height*sizeof(short));
      if (spixTmpImages[i].intermediaire_moyennes1 == NULL)
	 goto updateNotEnouhMemory;
      spixTmpImages[i].state_intermediaire_moyennes1 = SP_IMAGE_ALLOCATED;

      spixTmpImages[i].intermediaire_moyennes2 =
	malloc(width*height*sizeof(int));
      if (spixTmpImages[i].intermediaire_moyennes2 == NULL)
	 goto updateNotEnouhMemory;
      spixTmpImages[i].state_intermediaire_moyennes2 = SP_IMAGE_ALLOCATED;

      /*
       * les 2 images des moyennes et sigmas
       */

      spixTmpImages[i].im_moyennes = malloc(width*height*sizeof(float));
      if (spixTmpImages[i].im_moyennes == NULL)
	 goto updateNotEnouhMemory;
      spixTmpImages[i].state_moyennes = SP_IMAGE_ALLOCATED;

      spixTmpImages[i].im_sigmas = malloc(width*height*sizeof(float));
      if (spixTmpImages[i].im_sigmas == NULL)
	 goto updateNotEnouhMemory;
      spixTmpImages[i].state_sigmas = SP_IMAGE_ALLOCATED;


      spixTmpImages[i].state = SP_IMAGES_TMP_ALLOCATED;

   }

   return OK;

  updateNotEnouhMemory:

   *bilan = S_stereopixel_NOT_ENOUGH_MEMORY;

   for(i=0;i<2;i++) {
      if (spixTmpImages[i].intermediaire_moyennes1 != NULL)
	 free(spixTmpImages[i].intermediaire_moyennes1);
      if (spixTmpImages[i].intermediaire_moyennes2 != NULL)
	 free(spixTmpImages[i].intermediaire_moyennes2);
      if (spixTmpImages[i].im_moyennes != NULL)
	 free(spixTmpImages[i].im_moyennes);
      if (spixTmpImages[i].im_sigmas != NULL)
	 free(spixTmpImages[i].im_sigmas);
   }

   return ERROR;
}

static STATUS
spixUpdateCorrelResult(int *bilan)
{
   int    i;

   /* Variables déterminées */
   int    disp_max_general;
   int    score_tab_size;	/* (delta_max_disp * nbcol * sizeof(double)) */
   int    produits_tab_size;	/* (delta_max_disp * nbcol * sizeof(int)) */
   int    sommes_tab_size;	/* ( delta_max_disp * sizeof (int)) */

   if (spixCorrelResults.state == SP_CORRELATION_RESULTS_ALLOCATED)
      sc_clean_correlation_results(&spixCorrelResults);

   /*
    * On alloue les tableaux de valeurs min et max de la disparité
    */

   spixCorrelResults.tab_disp_min = malloc(width * sizeof(int));
   if (spixCorrelResults.tab_disp_min == NULL)
      goto notEnoughMem;

   spixCorrelResults.tab_disp_max = malloc(width * sizeof(int));
   if (spixCorrelResults.tab_disp_max == NULL)
      goto notEnoughMem;

   spixCorrelResults.tab_disp_nj2_min = malloc(width * sizeof(int));
   if (spixCorrelResults.tab_disp_nj2_min == NULL)
      goto notEnoughMem;

   spixCorrelResults.tab_disp_nj2_max = malloc(width * sizeof(int));
   if (spixCorrelResults.tab_disp_nj2_max == NULL)
      goto notEnoughMem;


   /*
    * On détermine les tableaux de disparité
    */

   calcule_disp_tableaux(width, spixCorrelParams.win_nb_col/2,
			 spixCorrelParams.disp_min,
			 spixCorrelParams.disp_max,
			 spixCorrelResults.tab_disp_max,
			 spixCorrelResults.tab_disp_min,
			 spixCorrelResults.tab_disp_nj2_max,
			 spixCorrelResults.tab_disp_nj2_min,
			 &disp_max_general,
			 &spixCorrelResults.disp_min_general);

   spixCorrelResults.delta_max_disp =
	disp_max_general - spixCorrelResults.disp_min_general;
   spixCorrelResults.win_nb_lig = spixCorrelParams.win_nb_lig;

   /*
    * Tailles des tableaux des scores
    */

   score_tab_size = spixCorrelResults.delta_max_disp * width * sizeof(double);
   produits_tab_size = spixCorrelResults.delta_max_disp * width * sizeof(int);
   sommes_tab_size = spixCorrelResults.delta_max_disp * sizeof(int);

   spixCorrelResults.tab_scores_2 = malloc(score_tab_size);
   if (spixCorrelResults.tab_scores_2 == NULL) goto notEnoughMem;

   spixCorrelResults.somme_pdt_bas = malloc(sommes_tab_size);
   if (spixCorrelResults.somme_pdt_bas == NULL) goto notEnoughMem;

   spixCorrelResults.somme_pdt_haut = malloc(sommes_tab_size);
   if (spixCorrelResults.somme_pdt_haut == NULL) goto notEnoughMem;

   spixCorrelResults.tab_tab_produits =
	malloc(spixCorrelParams.win_nb_lig*sizeof(int *));
   if (spixCorrelResults.tab_tab_produits == NULL) goto notEnoughMem;

   for(i=0; i<spixCorrelParams.win_nb_lig; i++) {
      spixCorrelResults.tab_tab_produits[i] = malloc(produits_tab_size);
      if (spixCorrelResults.tab_tab_produits[i] == NULL)
	 goto notEnoughMem;
   }

   spixCorrelResults.tab_produits_bas = malloc(produits_tab_size);
   if (spixCorrelResults.tab_produits_bas == NULL) goto notEnoughMem;

   spixCorrelResults.tab_scores = malloc(produits_tab_size);
   if (spixCorrelResults.tab_scores == NULL) goto notEnoughMem;

   /*
    * Suivent toutes les allocations des images résultats : l'image de
    * disparités...
    */

   spixCorrelResults.im_disparites = malloc(width*height * sizeof(float));
   if (spixCorrelResults.im_disparites == NULL) goto notEnoughMem;

   spixCorrelResults.state_disparites = SP_IMAGE_ALLOCATED;

   /*
    * l'image des écats types sur les disparités
    */

   spixCorrelResults.im_sigma_disparites = malloc(width*height * sizeof(float));
   if (spixCorrelResults.im_sigma_disparites == NULL) goto notEnoughMem;

   spixCorrelResults.state_sigma_disparites = SP_IMAGE_ALLOCATED;

   /*
    * l'image des régions.
    */

   spixCorrelResults.im_regions = malloc(width*height * sizeof(int));
   if (spixCorrelResults.im_regions == NULL) goto notEnoughMem;

   spixCorrelResults.state_regions = SP_IMAGE_ALLOCATED;
   spixCorrelResults.state = SP_CORRELATION_RESULTS_ALLOCATED;

   /*
    * Quelques tailles
    */

   spixCorrelResults.images_sizes.nb_lig_final = height;
   spixCorrelResults.images_sizes.nb_col_final = width;

   return OK;

  notEnoughMem:
   *bilan = S_stereopixel_NOT_ENOUGH_MEMORY;
   sc_clean_correlation_results(&spixCorrelResults);
   return ERROR;
}

static STATUS
spixUpdateIm3d(int *bilan)
{
   empty_data_im3d(&spixIm3d);

   if (init_data_im3d(&spixIm3d, height, width, 1/*covar*/, 0,0) != ERR_IM3D_OK) {
      *bilan = S_stereopixel_NOT_ENOUGH_MEMORY;

      empty_data_im3d(&spixIm3d);
      return ERROR;
   }

   return OK;
}


/* --- rgbmemcpy ----------------------------------------------------- */

/** Copy RGBA source image to intensity dst image.
 *
 * \param[out] dst
 */

static void
rgbmemcpy(unsigned char *dst, unsigned char *src, size_t n)
{
  unsigned char r, g, b;
  int i;

  for(i=0; i<n; i++) {
    r = *(src++); g = *(src++); b = *(src++);

    *(dst++) = 0.299 * r + 0.587 * g + 0.114 * b;
  }
}
